﻿using System;
using System.Linq;
using HeuristicLab.Services.Hive.Common;
using HeuristicLab.Services.Hive.Common.DataTransfer;

namespace HeuristicLab.Services.Hive {
  /// <summary>
  /// This class offers methods for cleaning up offline slaves and jobs
  /// </summary>
  public class LifecycleManager : ILifecycleManager {
    private DataAccess.IHiveDao dao {
      get { return ServiceLocator.Instance.HiveDao; }
    }
    private HeuristicLab.Services.Hive.DataAccess.TransactionManager trans {
      get { return ServiceLocator.Instance.TransactionManager; }
    }
    private IAuthorizationManager auth {
      get { return ServiceLocator.Instance.AuthorizationManager; }
    }
    private ILogger log {
      get { return LogFactory.GetLogger(this.GetType().Namespace); }
    }

    public void Cleanup() {
      log.Log("LifecycleManager.Cleanup()");
      SetTimeoutSlavesOffline();
      FinishParentJobs();
    }

    /// <summary>
    /// Searches for slaves which are timed out, puts them and their jobs offline
    /// </summary>
    private void SetTimeoutSlavesOffline() {
      var slaves = dao.GetSlaves(x => x.SlaveState != SlaveState.Offline);
      foreach (Slave slave in slaves) {
        if (!slave.LastHeartbeat.HasValue || (DateTime.Now - slave.LastHeartbeat.Value).TotalSeconds > ApplicationConstants.HeartbeatTimeout) {
          slave.SlaveState = SlaveState.Offline;
          SetJobsWaiting(slave.Id);
          dao.UpdateSlave(slave);
        }
      }
    }

    /// <summary>
    /// Looks for parent jobs which have FinishWhenChildJobsFinished and set their state to finished
    /// </summary>
    private void FinishParentJobs() {
      var parentJobsToFinish = dao.GetParentJobs(dao.GetResources(x => true).Select(x => x.Id), 0, true);
      foreach (var job in parentJobsToFinish) {
        dao.UpdateJobState(job.Id, JobState.Finished, null, null, string.Empty);
      }
    }

    private void SetJobsWaiting(Guid slaveId) {
      var jobs = dao.GetJobs(x => x.State == JobState.Calculating).Where(x => x.StateLog.Last().SlaveId == slaveId);
      foreach (var j in jobs) {
        j.StateLog.Add(new StateLog() { 
          State = JobState.Waiting,
          JobId = j.Id,
          DateTime = DateTime.Now,
          Exception = "Slave timed out"
        });
        dao.UpdateJob(j);
      }
    }

  }
}
